package ajedrez.server.comunicaciones;


import java.net.MalformedURLException;
import java.rmi.AlreadyBoundException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.server.UnicastRemoteObject;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import ajedrez.client.comunicaciones.IJugador;
import ajedrez.server.dominio.Tablero;
import ajedrez.server.persistencia.Metodos;
import ajedrez.server.presen.FVentanaServidor;


public class AjedrezServer extends UnicastRemoteObject implements IAjedrezServer {
	private static final long serialVersionUID = 1L;
	private FVentanaServidor interfazGUI;
	private String sIP;  // Datos de conexin
	private int nPuerto;
	private Hashtable<Integer, Tablero> htPartidasAjedrez;
	private Hashtable<String, IJugador> htJugadoresConectados;
	private Vector<String> vJugadoresDisponibles;
	private int nIdPartida = 1;
	
	
	public AjedrezServer() throws RemoteException {
		htPartidasAjedrez = new Hashtable<Integer, Tablero>();
		htJugadoresConectados = new Hashtable<String, IJugador>();
		vJugadoresDisponibles = new Vector<String>();
		nIdPartida = Metodos.getUltimoId()+1;
	} /* Fin constructor */
	
	
	/**
	 * Obtiene la instancia de FVentanaServidor que se comunica con AjedrezServer.
	 * @param interfazGUI   Instancia de la clase FVentanaServidor que se 
	 *                      comunica con AjedrezServer.
	 */
	public void setInterfaz(FVentanaServidor interfazGUI) {
		this.interfazGUI = interfazGUI;
	} /* Fin mtodo setInterfaz */
	
	
	/**
	 * Devuelve un Vector con los nombres de los jugadores conectados.
	 */
	public Vector<String> getJugadoresConectados() throws RemoteException {
		return vJugadoresDisponibles;
	} /* Fin mtodo getJugadoresConectados */
	
	
	/**
	 * Conecta el servidor a la IP y al puerto indicados.
	 */
	public void conectar(String sIP, int nPuerto) throws MalformedURLException, RemoteException {
		this.sIP = sIP;
		this.nPuerto = nPuerto;
		LocateRegistry.createRegistry(nPuerto);
		try {
			Naming.bind("rmi://"+sIP+":"+nPuerto+"/ServicioAjedrez", this);
		}
		catch (AlreadyBoundException eABE) {
			Naming.rebind("rmi://"+sIP+":"+nPuerto+"/ServicioAjedrez", this);
		}
	} /* Fin mtodo conectar */
	
	
	/**
	 * Desconecta todos los jugadores.
	 */
	public void desconectarJugadores() {
		// Se informa a todos los jugadores del cierre del servidor
		Enumeration<IJugador> eJugadores = htJugadoresConectados.elements();
		while (eJugadores.hasMoreElements()) {
			try {
				eJugadores.nextElement().cierreForzoso();
			}
			catch(RemoteException eR) {}
		}
	} /* Fin mtodo desconectarJugadores */
	
	
	/**
	 * Desconecta el servidor de la IP y del puerto indicados.
	 */
	public void desconectarServidor() throws RemoteException {
		// Se desconecta el servidor
		try {
			Naming.unbind("rmi://"+sIP+":"+nPuerto+"/ServicioAjedrez");
		}
		catch(MalformedURLException eMURL) {
			eMURL.printStackTrace();
		}
		catch(NotBoundException eNB) {
			eNB.printStackTrace();
		}
	} /* Fin mtodo desconectarServidor */
	
	
	/**
	 * Registra al usuario en la base de datos.
	 */
	public boolean registrar(String sUsuario, String sContrasea) throws RemoteException {
		boolean bRegistrado = Metodos.insertaJugador(sUsuario, sContrasea);
		if (bRegistrado) {
			// Se actualiza la ventana del servidor
			interfazGUI.aadeMensaje("El jugador " + sUsuario + " se ha registrado en el sistema.");
		}
		return bRegistrado;
	} /* Fin mtodo registrar */
	
	
	/**
	 * Comprueba la identificacin del usuario en la base de datos.
	 * @return   Devuelve '0' si la identificacin ha sido correcta, '-1' si
	 *           nicamente la contrasea est equivocada, '-2' si el usuario
	 *           no est registrado en el sistema y '-3' si el usuario ya se
	 *           encontraba identificado.
	 */
	public int identificar(String sUsuario, String sContrasea) throws RemoteException {
		int nValor = 0;
		// Se comprueba si el usuario ya est identificado
		if (htJugadoresConectados.containsKey(sUsuario)) {
			nValor = -3;
		}
		if (nValor != -3) {
			// Se valida al usuario en la base de datos
			nValor = Metodos.compruebaUsuario(sUsuario, sContrasea);
			if (nValor == 0) {
				//Se hace en identificar(String,IJugador)
			}
		}
		return nValor;
	} /* Fin mtodo identificar */
	
	
	/**
	 * Identifica al usuario en el sistema.
	 */
	public void identificar(String sUsuario, IJugador jugador) throws RemoteException {
		// Se informa al resto de jugadores de su llegada
		Enumeration<IJugador> eJugadores = htJugadoresConectados.elements();
		while (eJugadores.hasMoreElements()) {
			eJugadores.nextElement().mostrarJugadorNuevo(sUsuario);
		}
		// Se conecta al usuario al sistema
		htJugadoresConectados.put(sUsuario, jugador);
		vJugadoresDisponibles.addElement(sUsuario);
		// Se actualiza la ventana del servidor
		interfazGUI.aadeJugador(sUsuario);
		interfazGUI.aadeMensaje("El jugador " + sUsuario + " ha entrado en el sistema.");
	} /* Fin mtodo identificar */
	
	
	/**
	 * Finaliza la sesin del usuario.
	 * @return   '0' si se ha cerrado la sesin correctamente o '-1' si el
	 *           usuario no estaba identificado en el sistema.
	 */
	public int cerrarSesion(String sUsuario) throws RemoteException {
		int nValor = -1;
		if (htJugadoresConectados.containsKey(sUsuario)) {
			// Se desconecta al usuario del sistema
			htJugadoresConectados.remove(sUsuario);
			vJugadoresDisponibles.removeElement(sUsuario);
			nValor = 0;
			// Se informa al resto de jugadores de su salida
			Enumeration<IJugador> eJugadores = htJugadoresConectados.elements();
			while (eJugadores.hasMoreElements()) {
				eJugadores.nextElement().borrarJugSalaEspera(sUsuario);
			}
			// Se actualiza la ventana del servidor
			interfazGUI.eliminaJugador(sUsuario);
			interfazGUI.aadeMensaje("El jugador " + sUsuario + " ha salido del sistema.");
		}
		return nValor;
	} /* Fin mtodo cerrarSesion */
	
	
	/**
	 * Obtiene los datos clasificatorios del jugador.
	 * @return   Devuelve una matriz de cinco componentes:
	 *           0 - partidas jugadas
	 *           1 - partidas ganadas
	 *           2 - partidas perdidas
	 *           3 - partidas en tablas
	 *           4 - puntuacin
	 */
	public String[] getClasificacion(String sUsuario) throws RemoteException {
		return Metodos.getClasificacion(sUsuario);
	} /* Fin mtodo getClasificacion */
	
	
	/**
	 * Realizacin de un reto.
	 */
	public void retar(String sRetador, String sRetado) throws RemoteException {
		IJugador jug = htJugadoresConectados.get(sRetado);
		// Se informa al otro jugador de la peticin
		jug.retar(sRetador);
	} /* Fin mtodo retar */
	
	
	/**
	 * Recibe la contestacin del reto e informa al retador, creando la
	 * partida si es preciso.
	 */
	public void contestarReto(String sRetado, String sRetador, boolean bRespuesta) throws RemoteException {
		IJugador jugRetador = htJugadoresConectados.get(sRetador);
		if (bRespuesta) {  // Reto aceptado
			Tablero partida;
			// Identificador de la partida
			int nId = nIdPartida;
			nIdPartida++;
			// Se quitan los jugadores de la lista de disponibles
			vJugadoresDisponibles.removeElement(sRetador);
			vJugadoresDisponibles.removeElement(sRetado);
			// Se informa de la nueva lista de jugadores disponibles
			for (int i=0; i<vJugadoresDisponibles.size(); i++) {
				htJugadoresConectados.get(vJugadoresDisponibles.get(i)).borrarJugSalaEspera(sRetador);
				htJugadoresConectados.get(vJugadoresDisponibles.get(i)).borrarJugSalaEspera(sRetado);
			}
			// Sorteo de colores y creacin de la partida
			char cColorRetador;
			char cColorRetado;
			if (Math.random() < 0.5) {  // Retador juega con blancas
				partida = new Tablero(nId, sRetador, sRetado);
				cColorRetador = 'b';
				cColorRetado = 'n';
			}
			else {  // Retador juega con negras
				partida = new Tablero(nId, sRetado, sRetador);
				cColorRetador = 'n';
				cColorRetado = 'b';
			}
			// Almacenamiento temporal de la partida
			htPartidasAjedrez.put(nId, partida);
			// Se actualiza la ventana del servidor
			interfazGUI.aadeMensaje("Comienza la partida " + nId + ": " + partida.getJugador('b') + " - " + partida.getJugador('n'));
			// Se comunica a los dos jugadores su color y el identificador de la partida
			IJugador jugRetado = htJugadoresConectados.get(sRetado);
			jugRetado.contestarReto(nId, cColorRetado);
			jugRetador.contestarReto(nId, cColorRetador);
		}
		else {  // Reto rechazado
			jugRetador.contestarReto(-1, ' ');
		}
	} /* Fin mtodo contestarReto */
	
	
	/**
	 * Gestiona la peticin de un movimiento.
	 * @return   Clave indicativa del suceso producido. Puede ser:
	 *          0: el movimiento se ha realizado correctamente
	 *          1: estado 0 y el jugador actual ha eliminado una pieza del contrincante
	 *          2: estado 0 y la pieza es un pen que ha llegado al final del tablero
	 *          3: estados 1 y 2
	 *          4: estado 0 y el movimiento realizado ha sido tomar al paso
	 *          5: estado 0 y el movimiento realizado ha sido el enroque corto
	 *          6: estado 0 y el movimiento realizado ha sido el enroque largo
	 *         1_: estado menor de 10 y se produce jaque al rey contrario
	 *         2_: estado menor de 10 y el jugador actual ha ganado la partida
	 *         3_: estado menor de 10 y se han producido tablas
	 *         -1: no es el turno del jugador actual
	 *         -2: no hay ninguna pieza en el origen
	 *         -3: la pieza de la casilla origen no es del jugador actual
	 *         -4: hay una pieza del jugador actual en la casilla destino
	 *         -5: el movimiento no se corresponde con la pieza del origen
	 *         -6: hay una pieza en medio que impide el movimiento solicitado
	 *         -7: el movimiento dejara al rey en jaque
	 */
	public int mover(int nId, char cColor, String sCasillaOrigen, String sCasillaDestino) throws RemoteException {
		int [] anCoordenadas = convertirCoordenadas(sCasillaOrigen, sCasillaDestino);
				//[0]=filaOrigen, [1]=columnaOrigen, [2]=filaDestino, [3]=columnaDestino
		Tablero partida = findPartida(nId);
		int nCaso = partida.mover(cColor, anCoordenadas[0], anCoordenadas[1], anCoordenadas[2], anCoordenadas[3], sCasillaOrigen, sCasillaDestino);
		if (nCaso >= 0) {  // El movimiento es vlido
			//Se almacena el movimiento realizado
			String sMovimiento = "";
			if (nCaso%10 == 0) {  // Movimiento normal
				sMovimiento = sCasillaOrigen+"-"+sCasillaDestino;
			}
			else if (nCaso%10 == 5) {  // Enroque corto
				sMovimiento = "O-O";
			}
			else if (nCaso%10 == 6) {  // Enroque largo
				sMovimiento = "O-O-O";
			}
			else {  // Se ha comido una pieza
				sMovimiento = sCasillaOrigen+"x"+sCasillaDestino;
			}
			if (nCaso/10 == 1) {  // Ha habido jaque
				sMovimiento += " +";
			}
			else if (nCaso/10 == 2) {  // Ha habido jaque mate
				sMovimiento += " ++";
			}
			else {  // Ha habido tablas
			}
			partida.almacenarMovimiento(sMovimiento);
			if ((nCaso/10) > 1) {  // La partida ha terminado
				partida.guardarEnBD();
				// Se actualiza la ventana del servidor
				if (nCaso/10 == 2) {
					if (cColor == 'b') {
						interfazGUI.aadeMensaje("Finaliza la partida " + nId + ": " + "ha ganado el jugador " + partida.getJugador('b') + " con las blancas");
					}
					else {
						interfazGUI.aadeMensaje("Finaliza la partida " + nId + ": " + "ha ganado el jugador " + partida.getJugador('n') + " con las negras");
					}
				}
				else {  // (nCaso/10 == 3)
					interfazGUI.aadeMensaje("Finaliza la partida " + nId + ": " + "ha habido tablas");
				}
			}
			// Se informa al contrario del movimiento realizado
			String sContrario = "-";
			if (cColor == 'b') {
				sContrario = partida.getJugador('n');
			}
			else {
				sContrario = partida.getJugador('b');
			}
			IJugador jug = htJugadoresConectados.get(sContrario);
			jug.actualizar(nCaso,sCasillaOrigen,sCasillaDestino);
			if ((nCaso/10) > 1) {  // Se devuelven los jugadores a la lista de disponibles
				// Se informa de la nueva lista de jugadores disponibles
				for (int i=0; i<vJugadoresDisponibles.size(); i++) {
					htJugadoresConectados.get(vJugadoresDisponibles.get(i)).mostrarJugadorNuevo(partida.getJugador('n'));
					htJugadoresConectados.get(vJugadoresDisponibles.get(i)).mostrarJugadorNuevo(partida.getJugador('b'));
				}
				// Se aaden los jugadores a la lista de disponibles
				vJugadoresDisponibles.addElement(partida.getJugador('b'));
				vJugadoresDisponibles.addElement(partida.getJugador('n'));
			}
		}
		return nCaso;
	} /* Fin mtodo mover */
	
	
	/**
	 * Gestiona la conversin de un pen que ha llegado al final del tablero.
	 * @return   Clave indicativa del suceso producido. Puede ser:
	 *           0: conversin realizada sin incidencias
	 *           1: estado 0 y se produce jaque al rey contrario
	 *           2: estado 0 y el jugador actual ha ganado la partida
	 *           3: estado 0 y se han producido tablas
	 */
	public int convertirPeon(int nId, char cColor, char cTipo, int nFila, int nColumna) throws RemoteException {
		Tablero partida = findPartida(nId);
		int nCaso = partida.convertirPeon(cTipo);
		if (nCaso > 1) {  // La partida ha terminado
			partida.guardarEnBD();
			// Se actualiza la ventana del servidor
			if (nCaso/10 == 2) {
				if (cColor == 'b') {
					interfazGUI.aadeMensaje("Finaliza la partida " + nId + ": " + "ha ganado el jugador " + partida.getJugador('b') + " con las blancas");
				}
				else {
					interfazGUI.aadeMensaje("Finaliza la partida " + nId + ": " + "ha ganado el jugador " + partida.getJugador('n') + " con las negras");
				}
			}
			else {  // (nCaso/10 == 3)
				interfazGUI.aadeMensaje("Finaliza la partida " + nId + ": " + "ha habido tablas");
			}
		}
		// Se informa al contrario del cambio realizado
		String sContrario = "-";
		if (cColor == 'b') {
			sContrario = partida.getJugador('n');
		}
		else {
			sContrario = partida.getJugador('b');
		}
		IJugador jug = htJugadoresConectados.get(sContrario);
		jug.rivalCorono(nFila, nColumna, cTipo);
		if (nCaso > 1) {  // Se devuelven los jugadores a la lista de disponibles
			// Se informa de la nueva lista de jugadores disponibles
			for (int i=0; i<vJugadoresDisponibles.size(); i++) {
				htJugadoresConectados.get(vJugadoresDisponibles.get(i)).mostrarJugadorNuevo(partida.getJugador('n'));
				htJugadoresConectados.get(vJugadoresDisponibles.get(i)).mostrarJugadorNuevo(partida.getJugador('b'));
			}
			// Se aaden los jugadores a la lista de disponibles
			vJugadoresDisponibles.addElement(partida.getJugador('b'));
			vJugadoresDisponibles.addElement(partida.getJugador('n'));
		}
		return nCaso;
	} /* Fin mtodo convertirPeon */
	
	
	/**
	 * Un jugador ha propuesto tablas.
	 * @param nId      Identificador de la partida.
	 * @param cColor   Color del jugador.
	 */
	public void proponerTablas(int nId, char cColor) throws RemoteException {
		Tablero partida = findPartida(nId);
		String sContrario = "-";
		if (cColor == 'b') {
			sContrario = partida.getJugador('n');
		}
		else {
			sContrario = partida.getJugador('b');
		}
		// Se informa al otro jugador de la peticin
		IJugador jug = htJugadoresConectados.get(sContrario);
		jug.proponerTablas();
	} /* Fin mtodo proponerTablas */
	
	
	/**
	 * Recibe la contestacin de la propuesta de tablas e informa al jugador,
	 * finalizando la partida si es preciso.
	 * @param nId      Identificador de la partida.
	 * @param cColor   Color del jugador.
	 */
	public void contestarTablas(int nId, char cColor, boolean bRespuesta) throws RemoteException {
		Tablero partida = findPartida(nId);
		String sContrario = "-";
		if (cColor == 'b') {
			sContrario = partida.getJugador('n');
		}
		else {
			sContrario = partida.getJugador('b');
		}
		IJugador jug = htJugadoresConectados.get(sContrario);
		
		if (bRespuesta) {  //Tablas aceptadas
			// Se almacena la partida
			partida.guardarEnBD();
			// Se actualiza la ventana del servidor
			interfazGUI.aadeMensaje("Finaliza la partida " + nId + ": " + "los jugadores han acordado tablas");
			// Se informa al otro jugador
			jug.contestarTablas(true);
			// Se informa de la nueva lista de jugadores disponibles
			for (int i=0; i<vJugadoresDisponibles.size(); i++) {
				htJugadoresConectados.get(vJugadoresDisponibles.get(i)).mostrarJugadorNuevo(partida.getJugador('n'));
				htJugadoresConectados.get(vJugadoresDisponibles.get(i)).mostrarJugadorNuevo(partida.getJugador('b'));
			}
			// Se devuelven los jugadores a la lista de disponibles
			vJugadoresDisponibles.addElement(partida.getJugador('b'));
			vJugadoresDisponibles.addElement(partida.getJugador('n'));
		}
		else {  //Tablas rechazadas
			jug.contestarTablas(false);
		}
	} /* Fin mtodo contestarTablas */
	
	
	/**
	 * Un jugador ha solicitado abandonar.
	 * @param nId     Identificador de la partida.
	 * @param cColor  Color del jugador.
	 */
	public void abandonar(int nId, char cColor) throws RemoteException {
		Tablero partida = findPartida(nId);
		String sContrario = "-";
		if (cColor == 'b') {
			sContrario = partida.getJugador('n');
		}
		else {
			sContrario = partida.getJugador('b');
		}
		IJugador jug = htJugadoresConectados.get(sContrario);
		
		// Se almacena en y la partida
		partida.abandono(cColor);
		partida.guardarEnBD();
		// Se actualiza la ventana del servidor
		interfazGUI.aadeMensaje("Finaliza la partida " + nId + ": " + "el jugador " + partida.getJugador(cColor) + " ha abandonado");
		// Se informa al otro jugador
		jug.MostrarInfoAbandono();
		// Se informa de la nueva lista de jugadores disponibles
		for (int i=0; i<vJugadoresDisponibles.size(); i++) {
			htJugadoresConectados.get(vJugadoresDisponibles.get(i)).mostrarJugadorNuevo(partida.getJugador('n'));
			htJugadoresConectados.get(vJugadoresDisponibles.get(i)).mostrarJugadorNuevo(partida.getJugador('b'));
		}
		// Se devuelven los jugadores a la lista de disponibles
		vJugadoresDisponibles.addElement(partida.getJugador('b'));
		vJugadoresDisponibles.addElement(partida.getJugador('n'));
	} /* Fin mtodo abandonar */
	
	
	/**
	 * Busca la partida que se solicita.
	 * @param nId   Identificador de la partida.
	 * @return      Partida correspondiente al identificador.
	 */
	protected Tablero findPartida(int nId) throws RemoteException {
		return htPartidasAjedrez.get(nId);
	} /* Fin mtodo findPartida */
	
	
	/**
	 * Convierte las coordenadas reales (letra-nmero) a slo numricas.
	 */
	protected static int[] convertirCoordenadas(String sCasilla1, String sCasilla2) throws RemoteException {
		int [] anCoordenadas = new int[4];
				//[0]=filaOrigen, [1]=columnaOrigen, [2]=filaDestino, [3]=columnaDestino
		if (sCasilla1.substring(0,1).equals("a")) {
			anCoordenadas[1] = 8;
		}
		else if (sCasilla1.substring(0,1).equals("b")) {
			anCoordenadas[1] = 7;
		}
		else if (sCasilla1.substring(0,1).equals("c")) {
			anCoordenadas[1] = 6;
		}
		else if (sCasilla1.substring(0,1).equals("d")) {
			anCoordenadas[1] = 5;
		}
		else if (sCasilla1.substring(0,1).equals("e")) {
			anCoordenadas[1] = 4;
		}
		else if (sCasilla1.substring(0,1).equals("f")) {
			anCoordenadas[1] = 3;
		}
		else if (sCasilla1.substring(0,1).equals("g")) {
			anCoordenadas[1] = 2;
		}
		else if (sCasilla1.substring(0,1).equals("h")) {
			anCoordenadas[1] = 1;
		}
		anCoordenadas[0] = Integer.parseInt(sCasilla1.substring(1,2));
		if (sCasilla2.substring(0,1).equals("a")) {
			anCoordenadas[3] = 8;
		}
		else if (sCasilla2.substring(0,1).equals("b")) {
			anCoordenadas[3] = 7;
		}
		else if (sCasilla2.substring(0,1).equals("c")) {
			anCoordenadas[3] = 6;
		}
		else if (sCasilla2.substring(0,1).equals("d")) {
			anCoordenadas[3] = 5;
		}
		else if (sCasilla2.substring(0,1).equals("e")) {
			anCoordenadas[3] = 4;
		}
		else if (sCasilla2.substring(0,1).equals("f")) {
			anCoordenadas[3] = 3;
		}
		else if (sCasilla2.substring(0,1).equals("g")) {
			anCoordenadas[3] = 2;
		}
		else if (sCasilla2.substring(0,1).equals("h")) {
			anCoordenadas[3] = 1;
		}
		anCoordenadas[2] = Integer.parseInt(sCasilla2.substring(1,2));
		return anCoordenadas;
	} /* Fin mtodo convertirCoordenadas */
	
} /* Fin clase AjedrezServer */
